#if defined(VM_SCR) || defined(USE_VM)

#include <signal.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>

#include "QLtypes.h"
#include "QL68000.h"
#include "vm.h"
#include "unix.h"

#if 0
#ifdef GREGS
#include <sys/ucontext.h>
#include <sys/regset.h>
#endif
#endif 
 
/*                              UQLX                            
     catch VM errors that resulted from read/write of QDOS protected areas			     
*/

char *oldscr;
long pagesize;
int pageshift;


extern int main();
extern int strcasecmp();

char *scrModTable;
int sct_size;
int faultaddr;
int vm_ison=0;

int vmfatalretry=0;

#ifdef USE_VM
void segv_handler(int signr)
{
  fflush(stdout);
#if 0
    if (!inside_vmemu && !vm_ison)
    {
      printf("can't handle memory access error\n");
      abort();
    }
  if(!inside_vmemu && vm_ison)
    {
      vm_off();  /* maybe some driver did it, retry */
      return;
    }
#endif

  vm_undo_instruction();
  siglongjmp(instr_retry_catch,1);
}
#else
#if defined(VM_REGS) && defined(GREGS)
void segv_handler(int signr, siginfo_t info, ucontext_t *uap)
#else
#ifdef SPARC
void segv_handler(int signr, siginfo_t info)
#else
void segv_handler(int signr, siginfo_t *infop)
#endif
#endif
{
  int i,j;
#ifndef sgi
  siginfo_t *infop=&info; 
#endif
  
  if (signr!=SIGSEGV)
    {
      printf("wrong signal %d in SIGSEGV handler\n",signr);
      return;      
    }
  if (infop->si_code<0) return;

#if 0
  printf("address error - si_code %d\n",infop->si_code);
#endif


  switch(infop->si_code)
    {
    case SEGV_MAPERR : printf("no mapping\n");
    verbose_exit:
      printf("at address %x , QL relative %d\n",infop->si_addr,(Ptr)(infop->si_addr)-(Ptr)theROM);
      printf("faddr: %x\n",infop->si_addr);
      DbgInfo();
  
      abort();  /* needs some stackframe mungling */
      break;
    case SEGV_ACCERR : /*printf("wrong access mode\n");*/
      i=(int)((Ptr)infop->si_addr-(Ptr)theROM);
      if (i>=0x18000 && i<=0x18000+pagesize)
#ifndef VM_REGS
	printf("unexpected memory exception\n");
#else
	{
	  i-=0x18000;

	  if (vx)
	    printf("R/W hardaware reg: %x\n",i);
#if 0
	  DbgInfo();
#endif
	  if (i==0x21) /* a fairly complicated procedure here */
	    {     
	      if (retry_inst) /* instruction retry, it must be WRITE */
		{
		  retry_inst=0;
		  regEmuWrite=1;
		  if (vx)
		    printf("write IntReg\n");
		   
		  if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_WRITE )<0)
		    {
		      perror("fatal error: could not set memory protection");
		      abort();
		    }
		}
	      else /* try it with a read */
		{
		  retry_inst=1;
    
		  if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_READ | PROT_WRITE )<0)
		    {
		      perror("fatal error: could not set memory protection");
		      abort();
		    }
		  WriteByte(0x18020,0);
		  WriteByte(0x18021,theInt);
		  

		  if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_READ )<0)
		    {
		      perror("fatal error: could not set memory protection");
		      abort();
		    }
		}
	    }
	  
	  else  /* now the common case, registers we don't care reading/writing */
	    {
	      if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_READ | PROT_WRITE)<0)
		{
		  perror("fatal error: could not set memory protection");
		  abort();
		}
	  
	      WriteByte(0x18020,0);
	      WriteByte(0x18021,theInt);
	      GetDateTime(&j);
	      WriteLong(0x18000,j); 

	      if (vx)
		printf("theInt: %d\n",theInt);
	      theInt=0;
	      reInst=nInst;
	    }

/* set nInst=0: if we use GREGS that's not so easy: */
	  nInst=0;
#ifdef GREGS
#ifdef SPARC
          uap->uc_mcontext.gregs[REG_G7]=0;
#endif
#ifdef MIPS
#error "try something like the SPARC version here or just hope it will work" 
#endif
#ifdef HPPA
#error "try something like the SPARC version here or just hope it will work" 
#endif
#endif
	  regEmux=i;
#if 0
	  if (i!=0x21)
	    {
	      printf("Warning : hardware register read/write at %x may not ber supported\n",i+0x18000);
	      return;
	    }
#endif
	  return;
	}
#endif /*VM_REGS*/
      i-=qlscreen.qm_lo/*131072*/;
      if (i>=0 && i<=qlscreen.qm_len)
#ifndef VM_SCR
	printf("unexpected memory fault - VM_SCR not defined\n");
#else
	{
	  i=(w32)(i)/pagesize;
	  scrModTable[i]=1;
#if 0
	  printf("\tprocessing screen memory access %d\n",i); 
	  displayFrom=131072;  /* dummy !*/
	  displayTo=131072+100;
#endif	  
	  
	  if (mprotect((Ptr)theROM+qlscreen.qm_lo+i*pagesize,pagesize,PROT_READ | PROT_WRITE) <0)
	    perror("fatal failure in SIGSEGV handler - unable to change permission ");
	}
#endif /*VM_SCR*/
      else
	{
	  printf("unexpected memory fault\n");
	  goto verbose_exit;
	}
      break;
    default: printf("unknown reason for memory exception: %d\n",infop->si_code);
      break;
    }
#if 0
  printf("at address %x , QL relative %x\n",infop->si_addr,(Ptr)(infop->si_addr)-(Ptr)theROM);
  printf("faddr: %x",infop->si_addr);
#endif
}
#endif

#ifdef VM_SCR
void buserr_handler(int signr, siginfo_t info)
#else
void buserr_handler(int signr/*, siginfo_t info*/)
#endif
{
  if(signr!=SIGBUS)
    {
      printf("wrong signal %d in SIGBUS handler\n",signr);
      return;      
    }
  printf("bus error:\n");
  DbgInfo();
  abort();

#if 0
  switch(info.si_code)
    {
    case BUS_ADRALN: printf("wrong alignment\n");
      break;
    case BUS_ADRERR: printf("nonexistent address\n");
      break;
    default: printf("unknown reason\n");
      break;
    }
  printf("at address %x , QL relative %x\n",info.si_addr,(Ptr)info.si_addr-(Ptr)theROM);
  printf("faddr: %x",info.si_addr);
#endif
}

/* unlock memory in this region and mark it as changed */
void prepChangeMem(w32 from, w32 to)
{
  w32 lolx,hilx;
  int i;
  static int bsul=0;
  
  if (qlscreen.qm_lo==131072)
    {
      if (from<qlscreen.qm_lo+qlscreen.qm_len)
	{
	  lolx = from <qlscreen.qm_lo ? qlscreen.qm_lo : from;
	  hilx = to >qlscreen.qm_lo+qlscreen.qm_len ? qlscreen.qm_lo+qlscreen.qm_len : to;
	  
	  for(i=0;i<sct_size;i++) 
	    if (qlscreen.qm_lo+PAGEX(i)>=lolx &&
		qlscreen.qm_lo+PAGEX(i)<=hilx) scrModTable[i]=1;
	  
	  lolx= PAGEX(PAGEI(lolx));
	  hilx= PAGEX(PAGEI(hilx+pagesize-1));
	  
	  if (mprotect((Ptr)theROM+lolx, (hilx-lolx), PROT_READ | PROT_WRITE) <0)
	    {
	      perror("warning: unable to change VM permission, io to screen may fail ");
	    }
	}
      else /* extended screen */
	{
	  if (!bsul)
	    {
	      mprotect((Ptr)theROM+131072,32768, PROT_READ | PROT_WRITE);
	      bsul=1;
	    }
#if 0
	  if (to>qlscreen.qm_lo)
	    {
	      lolx = from <qlscreen.qm_lo ? qlscreen.qm_lo : from;
	      hilx = to >qlscreen.qm_hi ? qlscreen.qm_hi : to;
	      
	      for(i=0;i<sct_size;i++) 
		if (qlscreen.qm_lo+PAGEX(i)>=lolx &&
		qlscreen.qm_lo+PAGEX(i)<=hilx) scrModTable[i]=1;
	      
	      lolx= PAGEX(PAGEI(lolx));
	      hilx= PAGEX(PAGEI(hilx+pagesize-1));
	      
	      if (mprotect((Ptr)theROM+lolx, (hilx-lolx), PROT_READ | PROT_WRITE) <0)
		
		perror("warning: unable to change VM permission, io to screen may fail ");	      
	      
	    }
#endif
	}
    }
}

  
  
void vmMarkScreen(uw32 i)
{
  i=PAGEI((w32)(i-qlscreen.qm_lo));
  if (scrModTable[i]) return;
  
  scrModTable[i]=1;
#if 0
  printf("\tprocessing screen memory access %d\n",i); 
  displayFrom=131072;  /* dummy !*/
  displayTo=131072+100;
#endif	  
  
  if (mprotect((Ptr)theROM+qlscreen.qm_lo/*131072*/+i*pagesize,pagesize,PROT_READ | PROT_WRITE) <0)
    {
      perror("fatal failure in SIGSEGV handler - unable to change permission ");
      printf("pagesize %d, QL address %d, screen page number %d\n",pagesize,131072+i*pagesize,i);
      abort();
    }
}




void vm_init(void)
{
  int res;
  int scratch;
  int i;
  
#if 0
  printf("addr of:\t main \t%x\n\t\t strcasecmp \t%x\n\t\t read \t%x\n",main,strcasecmp,read);
#endif

#ifndef sgi
  res=sysconf(_SC_MEMORY_PROTECTION); 
  if (!res){
    printf("sorry, there seems to be no VM support\n recompile without USE_VM\n ");
    exit();
  }
#endif
  
  pagesize=sysconf(_SC_PAGESIZE);
  for(pageshift=0,i=pagesize;i;i>>=1,pageshift++);
  pageshift--;
  
  if ((1<<pageshift) != pagesize)
    {
      printf("your machine seems to have a very odd pagesize of %d bytes\n",pagesize);
      printf("please recompile with -DNO_PSHIFT added to your buildflags\n");
      exit(1);
    }
  else
    printf("page size %d, page shift %d\n",pagesize,pageshift);


  scratch=malloc(16*1024*1024);   /* alloc security gap */
  theROM=malloc(RTOP/*16*1024*1024*/+16*pagesize);
  if (!theROM)
    {
      printf("could not allocate space for QDOS memory!");
      abort();
    }

  i=((long long)theROM/pagesize);
  theROM=(w32 *)((i+4)*pagesize);

  
  free(scratch);
}

void vm_setscreen()
{   
  int i,xsize;
  
  xsize=PAGEX(PAGEI(qlscreen.qm_len+pagesize-1));

  /*sct_size=32768/pagesize;*/
  sct_size=PAGEI(xsize);
  
  scrModTable=malloc(sct_size+1);
  for (i=0 ; i<sct_size ; i++) scrModTable[i]=0;  /* hack, force init display */
 
  oldscr=malloc(xsize);
  memset(oldscr,255,xsize);    /* force first screen draw !*/
}


void vm_on(void)
{
  int res;
  int i,x=0;
  
#if 0
  printf("turn write protection on\n");
#endif
  
#ifdef VM_SCR
  /*for(i=0; i<sct_size; i++) if (scrModTable[i]) break;
  if(i>=sct_size) */
  if(mprotect((char*)theROM+qlscreen.qm_lo,qlscreen.qm_len,PROT_READ)<0)  /* ROM and screen */
    perror("sorry, could not enable screen write protection");
  /*else
    {
      if(mprotect((char*)theROM+131072,(i-1)*pagesize,PROT_READ)<0) 
	perror("sorry, could not enable write protection");
    }*/
#endif

#ifdef VM_REGS  
  if(mprotect((Ptr)theROM+0x18000,pagesize,PROT_NONE)<0) /* HWREGS */
    perror("sorry, could not enable read protection");
  /*mprotect(theROM+131072,RTOP-13072-32*1024,PROT_READ | PROT_WRITE);*/
#endif
  vm_ison=1;

#ifdef USE_VM
  if(mprotect((Ptr)theROM+0x18000,pagesize,PROT_NONE)<0) /* HWREGS */
    perror("sorry, could not enable HW register protection");
#endif
#if defined(USE_VM) && !defined(VM_SCR)
  for(i=0; i<sct_size; i++) 
    if(!scrModTable[i])
      if(mprotect((char*)theROM+131072+PAGEX(i),pagesize,PROT_READ)<0)  /* ROM and screen */
    perror("sorry, could not enable screen write protection");
#endif
}

void vm_off(void)
{
  if (mprotect(theROM,131072+32768,PROT_READ | PROT_WRITE)<0)
    perror("could not remove memory protection");
  
  if (qlscreen.qm_lo!=131072)
    if (mprotect(qlscreen.qm_lo,qlscreen.qm_len,PROT_READ | PROT_WRITE)<0)
      perror("could not remove memory protection");
  
  vm_ison=0;
}


#if 0
void vm_regemu()
{
  retry_inst=0;
  if (regEmuWrite)
    {
      if (regEmux!=0x21)
	printf("what ? you write hardware register %x\n",regEmux);
      if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_READ)<0)
	perror("warning: could not set memory protection, this may go wrong");

      intReg=ReadByte(0x18021);
      if (vx)
	printf("IntReg=%d\n",intReg);
      regEmuWrite=0;
    }
  
#if 1
  memset((Ptr)theROM+0x18004,0,200);
  /*
  WriteByte(0x18020,0);
  WriteByte(0x18021,0);
  */
#endif
  if (vx)
    printf("vm_regemu %x\n",regEmux);
#if 0
  DbgInfo();
#endif
  if (mprotect((Ptr)theROM+0x18000,pagesize,PROT_NONE)<0)
    perror("warning: could not set memory protection, this may go wrong");
  regEmux=0;
  reInst=0;
}
#endif




#endif /* USE_VM */
